#include <stdio.h>
#include <libaio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/time.h>

#include "util.h"

#define AIO_RING_MAGIC			0xa10a10a1
#define AIO_RING_COMPAT_FEATURES	1
#define AIO_RING_INCOMPAT_FEATURES	0
struct aio_ring {
	unsigned	id;	/* kernel internal index number */
	unsigned	nr;	/* number of io_events */
	unsigned	head;
	unsigned	tail;

	unsigned	magic;
	unsigned	compat_features;
	unsigned	incompat_features;
	unsigned	header_length;	/* size of aio_ring */


	struct io_event		io_events[0];
}; /* 128 bytes + ring size */


/* 
 * get an event from the event buffer.  returns 0 if successful and -1 if no event
 * available.
 */
int readevent (io_context_t ctx, struct io_event *event) {
    struct aio_ring *ring;
    ring = (struct aio_ring *)ctx;
    if (ring->head != ring->tail) {
        int head = ring->head;
        memcpy (event, &ring->io_events[head], sizeof (struct io_event));
        ring->head = (head+1) % ring->nr;
    } else 
        return -1;
    return 0;
}

#define BUFSIZE 1024
#define REPEAT (1024*1024)

int main (int argc, char **argv) {
    int fd, i, ret;
    io_context_t ctx;
    struct io_event events[1];
    struct iocb cb, *cbp = &cb;
    struct aio_ring *ring;
    char buf[BUFSIZE];
    struct timeval s1, f1, s2, f2;
    
    fd = open("/etc/services", O_RDONLY);
    if (fd <= 0) {
        perror("opening file");
        exit(-1);
    }
     
    memset(&ctx, 0 ,sizeof(ctx));
    if (io_setup(1024, &ctx) < 0) {
        perror ("setting up ioctx");
        exit (-1);
    }
    
    ring = (struct aio_ring *)ctx;

    printf("ID=%u, nr=%u, head=%u, tail=%u\n",
           ring->id,
           ring->nr,
           ring->head,
           ring->tail);
    printf("magic=%X, compat_features=%u, incompat_features=%u, header_length=%u\n",
           ring->magic,
           ring->compat_features,
           ring->incompat_features,
           ring->header_length);
    
    
    /* try read events from the buffer directly */
    if (read (fd, buf, sizeof(buf)) < 0) {
        perror("Reading file");
        exit (-1);
    }
    
    gettimeofday (&s1, 0);
    for (i = 0; i < REPEAT; i++) {
        io_prep_pread(cbp, fd, buf, sizeof(buf), 0);
        ret = io_submit (ctx, 1, &cbp);
        if (ret < 0) {
            perror ("submitting request");
            exit (-1);
        }
        if (io_getevents(ctx, 1, 1, events , NULL) < 0) {
            perror ("getting events");
            exit (-1);
        }
    }            
    gettimeofday (&f1, 0);
    
    gettimeofday (&s2, 0);
    for (i = 0; i < REPEAT; i++) {
        io_prep_pread(cbp, fd, buf, sizeof(buf), 0);
        ret = io_submit (ctx, 1, &cbp);
        if (ret < 0) {
            perror ("submitting request");
            exit (-1);
        }
        if (readevent(ctx, events) < 0) {
            perror ("aio request error");
            exit (-1);
        }
    }
    gettimeofday (&f2, 0);    

    printf ("Total time: %g\n", duration (&s1, &f2));
    printf ("io_submit+io_getevent: %g us, io_submit+readevent: %g us\n",
        duration (&s1, &f1) / REPEAT * 1000000,
        duration (&s2, &f2) / REPEAT * 1000000);

    return 0;
}
