/*
 * pagefault_pthread.c
 *
 * pthread counterpart of pagefault_coro3
 * read bits of files in parallel and compute checksum
 */

#include <stdio.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <sys/time.h>
#include <stdlib.h>
#include <pthread.h>
#include <string.h>
#include <signal.h>

/* Number of reads to perform */
int N = 100;
#define MAX_WORKER_NR 1000
#define MAX_N 10000

#define VERBOSE 0

pthread_t workers[MAX_WORKER_NR];
int nr_worker;
int rands[MAX_N];
char *data;     // all data that's read

struct worker_context {
    /* user stuff */
    int id;
    char *buffer;
    int length;
};

/* Worker coroutine that reads from a mmap'ed buffer */
void *worker(void *_ctx) {
    struct worker_context *ctx;
    int i, off;
    unsigned checksum = 0;
    char *pdata = data + ctx->id * N;

    /* Get context from main coroutine */
    ctx = (struct worker_context *)_ctx;

    for (i = 0; i < N; i++) {
        off = rands[i] % (ctx->length);
#if VERBOSE
        printf("%d.\n", ctx->id);
#endif
        checksum ^= (pdata[i] = ctx->buffer[off]);
    }

//    printf(".");
#if VERBOSE
    printf("\nCoroutine %d done\n", ctx->id);
#endif
    return (void *)checksum;
}

int main(int argc, char **argv) {
    int i;
    struct stat st;
//    struct sigaction sa;

    if (argc < 3) {
        printf("pagefault_pthread <number of reads per thread> <files>+\n");
        exit(0);
    }

    srand(0);
    
    N = atoi(argv[1]);
    if (N <= 0) {
      printf("N must > 0\n");
      exit(0);
    }
    for (i = 0; i < N; i++) {
        rands[i] = rand();
    }
    
    /* Create all workers */
    nr_worker = argc - 2;
    
    if (!(data = malloc(nr_worker * N))) {
        perror("Cannot allocate memory");
        exit(-1);
    }
    
    for (i = 0; i < nr_worker; i++) {
        struct worker_context *worker_context;
        int fd;

        worker_context = malloc(sizeof(struct worker_context));
        if (!worker_context) {
            perror("Error allocating worker context");
            exit(-1);
        }
        memset(worker_context, 0, sizeof(struct worker_context));
        
        worker_context->id = i;
        
        /* Open and mmap the file */
        fd = open(argv[i+2], O_RDONLY);
        if (fd < 0) {
            printf("Error opening file %s\n", argv[i+2]);
            perror("");
            exit(-1);
        }
        if (fstat(fd, &st)) {
            printf("Cannot stat file %s\n", argv[i+2]);
            perror("");
            exit(-1);
        }
        worker_context->buffer = mmap(0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
        worker_context->length = st.st_size;

        if (worker_context->buffer == MAP_FAILED) {
            printf ("Error mmapping %s\n", argv[i+1]);
            perror ("");
            exit(-1);
        }
        
        if (pthread_create(&workers[i], 0, worker, worker_context)) {
            printf("Cannot create thread %d\n", i);
            perror("");
            exit(-1);
        }

    }

    /* join all the threads */
    for (i = 0; i < nr_worker; i++) {
        int result;
        pthread_join(workers[i], (void **)&result);
        printf("checksum[%d] = %d\n", i, result);
    }

#if VERBOSE
    for (i = 0; i < nr_worker; i++) {
        printf("data[%d]=\n", i);
        for (j = 0; j < N; j++) {
            printf("%02x ", data[i*N+j]);
        }
        printf("\n\n");
    }
#endif

    return 0;
}
